from scipy.spatial import Delaunay
from numpy import transpose, array, ndarray
import numpy as np

def n_dim_cube(n, init_length, init_width):

    unit = [[ -0.5, -0.5, 0.5, 0.5], [ -0.5, 0.5, 0.5, -0.5]]

    if n != 2:
        for k in range(n-2):
            l = len(unit[0])
            for i in range(len(unit)):
                unit[i] = unit[i]+unit[i]
            unit.append( [-0.5]*l + [0.5]*l )
    
    for l in range(len(unit)):
        if l == 0:
            unit[l] = [element * init_length for element in unit[l]]
        else:
            unit[l] = [element * init_width for element in unit[l]]

    return transpose(array(unit))


class Cluster():
    # non shifted shape, easy to expand
    non_rotated_cube = None
    rotated_cube = None# shifted coor of rotated cube
    centroid = None# coor of center
    galaxies = None
    grow_limit = None
    new_galaxies_koef = None
    grow_function = None
    elongate_grow = None
    prev_n = None
    prev_v = None
    lr = None
    n_dim = 0
    times_grow = 0
    epsilon = 0
    '''
    center - coordinates of center of cluster (in case on sigle galaxy coordinates on galaxy)
    non_rotated_cube - coordinates of non rotated parallelepiped that fits to cluster
    galaxies - coordinates on galaxies
    init_length - ration on longer part of parallelepiped to shorters
    '''
    def __init__(self, center, non_rotated_cube = None,
                galaxies = None, epsilon = 0.5, n_dim = 3,
                grow_limit = 5, prev_n = None, prev_v = None,
                lr =1, elongate_grow = 1, grow_function = 'density'):
        
        self.was_complete = False
        
        self.n_dim = n_dim 
        self.elongate_grow = elongate_grow
        self.grow_limit = grow_limit
        self.prev_n = prev_n
        self.prev_v = prev_v
        self.lr = lr
        self.grow_function = grow_function

        self.non_rotated_cube = n_dim_cube(self.n_dim, epsilon, epsilon)
        
        if(isinstance(non_rotated_cube, ndarray)):
            self.non_rotated_cube = non_rotated_cube
        if(not isinstance(galaxies, ndarray)):
            self.galaxies = array([center])
        else:
            self.galaxies = array(galaxies)
        
        self.centroid = center[:self.n_dim]
        self.rotated_cube = self.rotate(self.centroid, self.non_rotated_cube)+self.centroid
        
        for gal in self.galaxies:
            if(Delaunay(self.rotated_cube).find_simplex(gal[:self.n_dim]) < 0):
                raise ValueError
        
    def get_length(self):
        length  =  self.non_rotated_cube[2, 0] - self.non_rotated_cube[0, 0]
        return length


    def get_width(self):
        width =  self.non_rotated_cube[1, 1] - self.non_rotated_cube[0, 0]
        return width

    def get_volume(self):
        volume = self.get_length()
        for i in range(self.n_dim-1):
            volume*=self.get_width()
        return volume

        
    def rotate(self, vector, points):    

        if(not np.any(vector)):
            return points

        x= array([1]+[0]*(self.n_dim-1))[np.newaxis, :].T
        y= array(vector)[np.newaxis, :].T
        y = y/np.linalg.norm(y)

        u = x/np.linalg.norm(x)
        v = y-(u.T@y)*u
        v = v/np.linalg.norm(v)

        cost = (x.T@y/np.linalg.norm(x)/np.linalg.norm(y))[0][0]
        sint = np.sqrt(1-np.power(cost,2))

        R = np.eye(len(x)) - u.dot(u.T) - v.dot(v.T)+(np.concatenate([u,v],axis = 1)@np.array([[cost, -sint],[sint,cost]]))@np.concatenate([u,v],axis = 1).T
        
        rotmat = []
        for p in points:
            rotmat.append(R@p)

        return array(rotmat)

    def isComplete(self):
        return self.times_grow == self.grow_limit

    def isWasComplete(self):
        return self.was_complete


    def grow(self):
        if(self.isComplete()):
            
            self.was_complete = True
            return
        self.times_grow += 1

        if(self.prev_n is None):
            composite_koef = np.power( self.lr  * (1 + 1/len(self.galaxies)),1/self.n_dim)
        else:
            if (self.grow_function == 'density'):
                nu1 = len(self.galaxies)*1.0/self.get_volume()
                nu2 = self.prev_n*1.0 / self.prev_v
                composite_koef = np.clip(np.power( self.lr*(len(self.galaxies)*1.0/self.prev_n) * (1 + nu2/nu1) ,1/self.n_dim), 1, 2)
            elif (self.grow_function == 'normal'):
                composite_koef = np.clip(np.power( self.lr*(len(self.galaxies)*1.0/self.prev_n) * (1 + 1/len(self.galaxies)) ,1/self.n_dim), 1, 2)
            else:
                raise ValueError('Incorrect grow_function value')
        self.prev_n = None
        self.prev_v = None
        self.non_rotated_cube[0] = self.non_rotated_cube[0]*composite_koef
        self.non_rotated_cube[1:,:] = self.non_rotated_cube[1:,:]*(1 + (composite_koef-1)/self.elongate_grow)        
        self.rotated_cube = self.rotate(self.centroid, self.non_rotated_cube)+self.centroid         


import plotly.graph_objects as go
def go_cluster(cluster):
    
    if(len(cluster.galaxies)==1):
        return [go.Scatter3d(
    x=cluster.galaxies[:,0],
    y=cluster.galaxies[:,1],
    z=cluster.galaxies[:,2],
    mode='markers',
    marker=dict(
        size=1,
        opacity=1,
        color='green'
    )
    )]
    
    return [go.Scatter3d(
    x=cluster.galaxies[:,0],
    y=cluster.galaxies[:,1],
    z=cluster.galaxies[:,2],
    mode='markers',
    marker=dict(
        size=3,
        opacity=1,
        color='green'

    )
    )]+[get_mesh(cluster)]+[go.Scatter3d(
    x=[cluster.centroid[0]],
    y=[cluster.centroid[1]],
    z=[cluster.centroid[2]],
    mode='markers',
    marker=dict(
        size=2,
        opacity=1,
        color='red'
    )
    )]

def get_mesh(self):
        return go.Mesh3d(
        x=self.rotated_cube[:,0],
        y=self.rotated_cube[:,1],
        z=self.rotated_cube[:,2],
        opacity=0.5,
        i = [7, 0, 0, 0, 4, 4, 6, 6, 4, 0, 3, 2],
        j = [3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3],
        k = [0, 7, 2, 3, 6, 7, 1, 1, 5, 5, 7, 6],color='blue'
        )

def get_clusters_plot(clusters,is_show=False,filename='3d_model'):
    
    data=[]
    for cluster in clusters:
        data+=go_cluster(cluster)
    
    fig = go.Figure(data=data)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0))
    if(is_show):
        fig.show()
    fig.write_html(filename+".html")


# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name='pbgca',
    packages=find_packages(include=['pbgca']),
    version='0.1.0',
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    description="Physically based galaxy clustering algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jeromeshan/pbgca",
    author="Andrey Shan",
    author_email="avshan@edu.hse.ru",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    include_package_data=True,
    install_requires=["numpy","pandas","ray","scipy"]
)


import re
from typing import List
import ray
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
from scipy.spatial import Delaunay
from scipy.spatial.distance import cdist
import time
from numpy import transpose, array, arange, concatenate, eye, unique, dot, zeros, append, ones
import numpy as np
from numpy.linalg import norm
from .cluster import Cluster, n_dim_cube
from scipy.spatial import distance

class ConnectivityMatrix:
    def __init__(self, clusters:List[Cluster],epsilon,limit_radian,n_dim,parallel):

        self.data = []
        self.epsilon = epsilon
        self.limit_radian = limit_radian
        self.parallel = parallel
        for cluster in clusters:
            self.data.append([cluster.centroid.reshape(1,-1),#0
            len(cluster.galaxies),#1
            cluster.get_length(),#2
            cluster.rotated_cube,#3
            cluster.galaxies[:,:n_dim],
            cluster.isWasComplete()
            ])
        self.data = np.array(self.data,dtype=object)
    
    def split_array(self, a, n):
        k, m = divmod(len(a), n)
        return (a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))

    def get_matrix(self):
        if self.parallel:
            data = ray.put(self.data)
            graph = ray.get([self.parallel_connectivity_matrix.remote(self, data, subset) for subset in self.split_array(arange(len(self.data)), 64)])
            graph = concatenate( graph, axis = 0 )
        else:
            graph = self.connectivity_matrix(self.data, arange(len(self.data)))

        graph = graph+transpose(graph)
        graph = graph-eye(len(self.data))
        return graph


    def connectivity_matrix(self, data , rows):
        new_graph = zeros((len(rows), len(data)))
        for i in range(len(rows)):
            for j in range(rows[i], len(data)):

                if(rows[i] == j):
                    new_graph[i, j] = 1
                    continue

                halfsum_len = (data[rows[i],2]+data[j,2])/2
                centroid_i = data[rows[i],0]
                centroid_j = data[j,0]
                n_dim = len(centroid_i)
                cube_i = data[rows[i],3]
                cube_j = data[j,3]
                p_i = data[rows[i],4]
                p_j = data[j,4]
                
                if(data[rows[i],5] and data[j,5]):
                    continue
                
                if((data[rows[i],1]==1) and (data[j,1]==1)):
                    if(cdist(centroid_i,centroid_j,'euclidean')<self.epsilon):
                        new_graph[i, j] = 1
                    continue

                centroids_diff =centroid_i[0]-centroid_j[0]
                dim_level_check = True
                for k in range(n_dim):
                    if(halfsum_len<np.abs(centroids_diff[k])):
                        dim_level_check = False
                        continue
                if(not dim_level_check):
                    continue


                if(np.arccos(1- cdist(centroid_i,centroid_j,'cosine')) > self.limit_radian):
                    continue
 
                if(distance.cdist(p_i,p_j).min()>10*self.epsilon):
                    continue

                if(self.check_collision(cube_i,p_j)):
                    new_graph[i, j] = 1
                elif(self.check_collision(cube_j, p_i)):
                    new_graph[i, j] = 1
        return new_graph

    @ray.remote
    def parallel_connectivity_matrix(self, data, rows):
        return self.connectivity_matrix(data, rows)
        
    def check_collision(self, cube, p):
        delaunay = Delaunay(cube)
        for gal in p:
            if(delaunay.find_simplex(gal) >= 0):
                return True
        return False

    

class Clusterer:

    def __init__(self, epsilon = 0.5, lr = 1, max_iter = 30,
                limit_radian = 0.01, grow_limit = 3, elongate_grow = 1,
                grow_function = 'density', min_diff = 1, parallel = False):
        self.clusters = []
        self.epsilon = epsilon
        self.lr = lr
        self.max_iter = max_iter
        self.limit_radian = limit_radian
        self.grow_limit = grow_limit
        self.elongate_grow = elongate_grow
        self.grow_function = grow_function
        self.min_diff = min_diff
        self.parallel = parallel
 
    def merge(self):

        matrix_gen = ConnectivityMatrix(self.clusters,self.epsilon,self.limit_radian,self.n_dim,self.parallel)
        graph = matrix_gen.get_matrix()     

        graph = graph+transpose(graph)
        graph = graph-eye(len(self.clusters))
        graph = csr_matrix(graph)
        _, labels = connected_components(csgraph = graph, directed = False, return_labels = True)
        components = []
 
        np_clusters = array(self.clusters)
        components = [np_clusters[labels == label]  for label in unique(labels) ]
        new_clusters = list(map(self.collide_clusters, components))

        self.clusters = new_clusters        
 
    def collide_clusters(self, clusters:List[Cluster]):
 
        if(len(clusters) == 1):
            return clusters[0]

        max_prev_cluster = max(clusters, key = lambda c: len(c.galaxies))
        galaxies = [cluster.galaxies for cluster in clusters]
        vertex_points = [cluster.rotated_cube for cluster in clusters]
 
        galaxies = concatenate( galaxies, axis = 0 )
        vertex_points = concatenate( vertex_points, axis = 0 )
 
        center = galaxies[:, :self.n_dim].mean(axis = 0)
 
        projections = center * dot(galaxies[:, :self.n_dim], transpose([center])) / dot(center, center)
 
        distances_on_line = norm(projections-center, axis = 1)
        vectors_from_line = galaxies[:, :self.n_dim]-projections
 
        length = distances_on_line.max()
        width = norm(vectors_from_line, axis = 1).max()

        cube = n_dim_cube(self.n_dim, length*2+self.epsilon/2, width*2+self.epsilon/2)
 
        return Cluster(center, cube, galaxies,  n_dim = self.n_dim, prev_n= len(max_prev_cluster.galaxies),prev_v=max_prev_cluster.get_volume(), grow_limit = self.grow_limit, lr = self.lr, elongate_grow = self.elongate_grow, grow_function=self.grow_function)
 
    def compress_cluster(self, cluster):
 
        galaxies = cluster.galaxies
        if(galaxies.ndim == 1 ):
            galaxies = array([galaxies])
 
        center = galaxies[:, :self.n_dim].mean(axis = 0)
 
        projections = center * dot(galaxies[:, :self.n_dim], transpose([center])) / dot(center, center)
 
        distances_on_line = norm(projections-center, axis = 1)
        vectors_from_line = galaxies[:, :self.n_dim]-projections

        length = distances_on_line.max()
        width = norm(vectors_from_line, axis = 1).max()
 
        cube = n_dim_cube(self.n_dim, length*2+self.epsilon/10, width*2+self.epsilon/10)
        
        return Cluster(center, cube, galaxies, n_dim = self.n_dim)
 
    def step(self):
 
        # start = time.time()
        is_done = True
        for cluster in self.clusters:
            if(not cluster.isComplete()):
                is_done = False
                break
        # print('\t','complete check time: ', time.time()-start, ' s')

        if (is_done):
            return False
        
        
        if(not self.isFirstStep):
            # start = time.time()
            for cluster in self.clusters:
                cluster.grow()
            # print('\t','grow time: ', time.time()-start, ' s')
        else:
            self.isFirstStep = False

        # start = time.time()
        self.merge()
        # print('s\t','merge time: ', time.time()-start, ' s')

        return True

    def split_array(self, a, n):
        k, m = divmod(len(a), n)
        return (a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))
 

    def fit(self, data):

        if(self.parallel):
            if ray.is_initialized():
                ray.shutdown()
            ray.init()

        data_np = array(data)
        self.n_dim = len(data_np[0])
        self.clusters = [Cluster(append(data_np[i], i), epsilon = self.epsilon, n_dim = self.n_dim,lr = self.lr,elongate_grow = self.elongate_grow, grow_function=self.grow_function) for i in range(len(data_np)) ]
        iter_num = 1
        self.isFirstStep = True
        start = time.time()
        prev_clusters_number = len(self.clusters)
        while(self.step()):
            print('iter : ', iter_num, ', n_clusters: ', len(self.clusters), ', time: ', time.time()-start, ' s')
            iter_num += 1
            start = time.time()

            if ((self.min_diff>0) and (prev_clusters_number - len(self.clusters) <= self.min_diff)):
                break
            else:
                prev_clusters_number = len(self.clusters)
            
            if(iter_num > self.max_iter):
                break

        galaxies = []
        galaxies = [append(self.clusters[i].galaxies, ones((len(self.clusters[i].galaxies), 1))*i , axis = 1) for i in range(len(self.clusters))]
        self.clusters = [self.compress_cluster(cluster) for cluster in self.clusters]
        galaxies = concatenate(galaxies)
        galaxies = galaxies[galaxies[:, self.n_dim].argsort()]  

        if(self.parallel):
            ray.shutdown()
        
        self.labels_ = galaxies[:, -1]

        return self.labels_